@@ -2,15 +2,33 @@
 
 import sys
 
-from stuf import orderedstuf, chainstuf
-# from otherstuf import chainstuf
+from stuf import orderedstuf  #, chainstuf
+# temporarily pulled stuf version of chainstuf in favor of
+# local version until a bug with generator values can be
+# tracked down and fixed
+from options.chainstuf import chainstuf
 from options.configparser import ConfigParser
-from options.nulltype import NullType
+from nulltype import NullType
 from options.funclike import *
+from options.combomethod import combomethod
 
+
+# Define sentinel objects
 Unset      = NullType('Unset')
 Prohibited = NullType('Prohibited')
 Transient  = NullType('Transient')
+Reserved   = NullType('Reserved')
+
+def bad_option_name(name):
+    return hasattr(Options, name) or hasattr(OptionsChain, name)
+
+
+class BadOptionName(AttributeError):
+    """
+    Raised when an Options object is given an option name that conflicts
+    with an existing method.
+    """
+    pass
 
 
 def attrs(m, first=[], underscores=False):
@@ -29,6 +47,35 @@ def attrs(m, first=[], underscores=False):
     return ', '.join(["{0}={1}".format(k, repr(m[k])) for k in keys])
 
 
+class OptionsClass(object):
+
+    """
+    Class from which client classes can inherit. It will add ``set`` and
+    ``settings`` methods.
+    """
+
+    @combomethod
+    def set(receiver, **kwargs):
+        """
+        Change the receiver's settings to those defined in the kwargs.
+        An update-like function. This uplevels calls that would look
+        like ``Class.options.set(...)`` to the simpler ``Class.set(...)``.
+        Works on either class or instance receivers. Requires that one
+        uses the instance variable ``options`` to store persistent
+        configuration data.
+        """
+        receiver.options.set(**kwargs)
+
+    def settings(self, **kwargs):
+        """
+        Open a context manager for a `with` statement. Temporarily change settings
+        for the duration of the with.
+        """
+        return OptionsContext(self, kwargs)
+
+    def __repr__(self):
+        return "{0}({1})".format(self.__class__.__name__, attrs(self.options))
+
 class Options(orderedstuf):
 
     """
@@ -38,17 +85,19 @@ class Options(orderedstuf):
     def __init__(self, *args, **kwargs):
         orderedstuf.__init__(self, *args)
         for k, v in kwargs.items():
-            self[k] = v
+            if bad_option_name(k):
+                raise BadOptionName("Options already defines an attribute {0!r}".format(k))
+            else:
+                self[k] = v
         self._magic = {}
 
         # The explicit for loop setting values from kwargs should not be necessary.
         # orderedstuf.__init__(self, *args, **kwargs) should do the trick. And it
         # does, on most platforms. But on PyPy, stuf(a=sys.stdout) fails in much
         # the same way used to on Python 3. Until stuf/orderedstuf is fixed for
-        # PyPy, this workaround fixes the issue for Options.
-
-    def __repr__(self):
-        return "{0}({1})".format(self.__class__.__name__, attrs(self))
+        # PyPy, this workaround fixes the issue for Options. (Reconfirmed bug
+        # still exists in PyPy 2.1 with stuf 0.9.12 in Sept 2013. Patch
+        # retained.)
 
     def set(self, **kwargs):
         # To set values at the base options level, create a temporary next level,
@@ -58,6 +107,46 @@ class Options(orderedstuf):
         for key in self.keys():
             self[key] = oc[key]
 
+    def setdefault(self, key, default):
+        """
+        If key is in the dictionary, return its value. If not, insert key with a
+        value of default and return default. If the value is Prohibited, no
+        change will be allowed, and a KeyError will be raised. If the value is
+        Transient or Unset, it will be as if the value were never there, and the
+        default will be set. NB ``setdefault`` does not participate in "magical"
+        value interpretation.
+        """
+        try:
+            value = self[key]
+            if value is Prohibited:
+                raise KeyError("changes to {0!r} are prohibited".format(key))
+            elif value is Transient or value is Unset:
+                self[key] = default
+                return default
+            else:
+                return value
+        except KeyError as e:
+            # All keys must be defined in initial object construction,
+            # even if defined as Unset or Transient. So if *no* value
+            # is available for a key, definite error. Stricter than
+            # general dict / mapping practice.
+            raise e
+
+        # In general usage, ``setdefault`` will be called only on ``OptionsChain``
+        # instances, but defined for ``Options`` as well for completeness.
+
+    def update(self, *args, **kwargs):
+        """
+        Update self with the given data. Necessary to avoid apparent bug in
+        ``stuf`` when updating with iterable objects, especially generators. NB
+        ``update`` does not participate in "magical" value interpretation.
+        """
+        for a in args:
+            for k, v in a.items():
+                self[k] = v
+        for k, v in kwargs.items():
+            self[k] = v
+
     def push(self, kwargs):
         """
         Create the next layer down. Intended for instances to call during
@@ -88,9 +177,15 @@ class Options(orderedstuf):
         """
         child = OptionsChain(self, {})
         for key, value in kwargs.items():
-            child[key] = value
+            if bad_option_name(key):
+                raise BadOptionName("OptionsChain already defines an attribute {0!r}".format(key))
+            else:
+                child[key] = value
         return child
 
+    def parent(self, new_base):
+        raise ValueError("Cannot change parent of base Options")
+
     def magic(self, **kwargs):
         """
         Set some options as having 'magical' update properties. In a sense, this
@@ -179,7 +274,8 @@ class OptionsChain(chainstuf):
         elif argcount == 3:
             return magicfn(None, value, self)
         else:
-            raise ValueError('magic function should have 1-3 arguments, not {0}'.format(argcount))
+            msg = 'magic functions take 1-3 arguments, not {0}'.format(argcount)
+            raise ValueError(msg)
 
     def _process(self, base, kwargs):
         """
@@ -218,7 +314,7 @@ class OptionsChain(chainstuf):
         guts = attrs(self, first=list(grandpa.keys()), underscores=True)
         return "{0}({1} layers: {2})".format(self.__class__.__name__, n_layers, guts)
 
-    #def __iter__(self):
+    # def __iter__(self):
     #    for k in super(OptionsChain, self).keys():
     #        if not k.startswith('_'):
     #            yield k
@@ -266,9 +362,19 @@ class OptionsChain(chainstuf):
         """
         child = OptionsChain(self, {})
         for key, value in kwargs.items():
-            child[key] = value
+            if bad_option_name(key):
+                raise BadOptionName("OptionsChain already defines an attribute {0!r}".format(key))
+            else:
+                child[key] = value
         return child
 
+    def parent(self, new_base):
+        """
+        Rebase the mapping (change its parent). Useful for implementing
+        "styles"
+        """
+        raise NotImplementedError()
+
     def set(self, **kwargs):
         newopts = self._process(self, kwargs)
         for k, v in newopts.items():
@@ -279,26 +385,44 @@ class OptionsChain(chainstuf):
             else:
                 self.maps[0][k] = v
 
-    def setdefault(self, key, default=None):
+    def setdefault(self, key, default):
         """
-        If key is in the dictionary, return its value. If not,
-        insert key with a value of default and return default.
-        default defaults to None. If the value is Prohibited,
-        no change will be allowed, and a KeyError will be raised.
-        If the value is Transient, it will be as if the value
-        were never there, and the default will be set.
+        If key is in the dictionary, return its value. If not, insert key with a
+        value of default and return default. If the value is Prohibited, no
+        change will be allowed, and a KeyError will be raised. If the value is
+        Transient or Unset, it will be as if the value were never there, and the
+        default will be set. NB ``setdefault`` does not participate in "magical"
+        value interpretation.
         """
         try:
             value = self[key]
             if value is Prohibited:
-                raise KeyError("changes to '{0}' are prohibited".format(key))
-            elif value is Transient:
+                raise KeyError("changes to {0!r} are prohibited".format(key))
+            elif value is Transient or value is Unset:
                 self[key] = default
                 return default
-            return value
+            else:
+                return value
         except KeyError as e:
+            # All keys must be defined in initial object construction,
+            # even if defined as Unset or Transient. So if *no* value
+            # is available for a key, definite error. Stricter than
+            # general dict / mapping practice.
             raise e
 
+    def update(self, *args, **kwargs):
+        """
+        Update self with the given data. Necessary to avoid apparent bug in
+        ``stuf``, ``orderedstuf``, or ``chainstuf`` when updating with iterable
+        objects, especially generators. NB ``update`` does not participate in "magical"
+        value interpretation.
+        """
+        for a in args:
+            for k, v in a.items():
+                self[k] = v
+        for k, v in kwargs.items():
+            self[k] = v
+
     def copy(self):
         """
         Return a copy of the self. That means a copy of just the top layer,
@@ -320,8 +444,7 @@ class OptionsChain(chainstuf):
     #  TBD Handle the unsetting of magic in subclasses
 
 
-
-    #def __setattr__(self, name, value):
+    # def __setattr__(self, name, value):
     #    print "OptionsChain.__setattr__() name:", name, "value:", value
     #    if name in self:
     #        print "    in self"
